// Represents a 3D object
// Author: Dan Fleck

#include "Object3d.h"
#include "DarkGDK.h"
#include "Matrix3.h"

Object3d::Object3d() {
	// Default constructor.... do nothing
}

void Object3d::setup(int id, int xL, int yL, int zL) {
	this->id = id;

	// Angular velocity to 0
	w[0] = 0; w[1]=0; w[2]=0;

	// Init forces to 0
	for (int i=0;i<3;i++) {
		totalForce[i] = 0;
		totalTorque[i] = 0;
	}
	this->tempCounter = 0; // DEBUG

	// Init orientation
	o[0] = 0; o[1] = 0; o[2] = 0;

	// Offset from CM where the point is
	pointLoc[0] = xL;
	pointLoc[1] = yL;
	pointLoc[2] = zL;

	// Init stuff

	// Dimensions of "total" box
	this->width = 70;
	this->height = 70;
	this->depth = 70;

	this->r[0] = 0; // X position
	this->r[1] = 0; // Y position
	this->r[2] = 0;  // Z position

	this->v[0] = 0; // Velocity of Center of Mass (X)
	this->v[1] = 0; // Velocity of Center of Mass (Y)
	this->v[2] = 0; // Velocity of Center of Mass (Z)

	this->A = new Matrix3(); // Orientation (Identity)

	this->M = 10; // Total Mass 

	this->L[0] = 0; // Angular Momentum (X)
	this->L[1] = 0; // Angular Momentum (Y)
	this->L[2] = 0; // Angular Momentum (Z)

	// Inverse inertia tensor
	this->Iinv = new Matrix3();
	this->createInertiaTensor(this->Iinv, width, height, depth, M);

	this->Icm = new Matrix3(); // Inertia in world coordinates
	

	// Compute initial auxiliary values
	this->computeIcm();
	this->computeAngularVelocity();

	// Make some rotation matrices -- this is really debugging code
	float tc = dbCOS(3);
	float ts = dbSIN(3);
	float tl[9] = { tc, 0, ts, 
					0 , 1, 0 ,
					-ts, 0, tc };
	tc = dbCOS(-3);
	ts = dbSIN(-3);
	float tr[9] = { tc, 0, ts, 
					0 , 1, 0 ,
					-ts, 0, tc };
	this->rotLeftM = new Matrix3(tl);
	this->rotRightM = new Matrix3(tr);
						 

	// Make a 3D object
	dbMakeObjectSphere(id, 5);

	// position the object in a random location
	dbPositionObject ( id, this->r[0], this->r[1], this->r[2] );

	// give the object a color
	dbColorObject ( id, dbRgb ( 200, 0, 0) );

	// increase specular power
	dbSetObjectSpecularPower ( id, 255 );

	// turn off ambient lighting for this object
	dbSetObjectAmbient ( id, 0 );
}

void Object3d::updatePhysics() {
	float h = 1.0f;

	// Integrate Quantities
	
	   // Update position
	   for (int i=0;i<3;i++) {
	    	r[i] = r[i] + h*v[i];
			if (dbABS(r[i]) < 0.0001) r[i] = 0;
	   }
	   // Update velocity
	   for (int i=0;i<3;i++) v[i] = v[i] + h*(totalForce[i]/M);

	   // Update orientation
	   updateOrientation(A, h, w);

	   // Update momentum
	   for (int i=0; i<3; i++) L[i] = L[i] + h*totalTorque[i];

	// Reorthogonalize A
	   A->OrthonormalizeOrientation();
	   
	// Compute auxiliary quantities
	   	this->computeIcm();
		this->computeAngularVelocity();

	
	// Reset total forces
		for (int i=0; i<3; i++) {
			totalForce[i] = 0;
			totalTorque[i] = 0;
		}
}

// Update object position and orientation
void Object3d::updatePosition() {
		float newOrientation[3];
		rotateObject(newOrientation);
		dbPositionObject(this->id, r[0]+newOrientation[0],r[1]+newOrientation[1],r[2]+newOrientation[2]);

}

// Rotate the vector representation of this "point"
void Object3d::rotateObject(float newOrientation[3]) {
	// newOrientation = A*pointLoc;
	A->multiply(pointLoc, newOrientation);
	
}

// Integrate the orientation
void Object3d::updateOrientation(Matrix3 * A, float h, float w[3]) {

	// Do tilde operator (so we can essentially cross product (w*h) X A
	float tempWM[9] = {0,	   -w[2]*h, w[1]*h, 
					   w[2]*h, 0,       -w[0]*h,
					  -w[1]*h, w[0]*h,  0 };

	Matrix3 *w3 = new Matrix3(tempWM);

	// w3 = w3 * A
	w3->multiply(A);

	// Now, add A to w3 and store in A
	A->add(w3);
	delete w3;
}


void Object3d::addForce(float force[3], int x, int y, int z) {
	
	// Create an impulse force given the location 
	// of the mouse, the direction (always -Z) and the 
	// magnitude, currently a constant

	// Compute the torque force
	float r[3] = { x, y, z};
	float torque[3]; 
	cross(r, force, torque);

	for (int i=0; i<3; i++) {
		//totalForce[i] = totalForce[i] + force[i];
		totalTorque[i] = totalTorque[i] + torque[i];
	}
}

// Compute current angular velocity
void Object3d::computeAngularVelocity() {
	
	// Multiply Icm by L and store in w
	this->Icm->multiply(L, w);
}

// Compute the Intertia tensor in world coordinates
// by multiplying by A, AT
void Object3d::computeIcm() {

	// Icm = A*Iinv
	this->A->multiply(this->Iinv, this->Icm);
	
	// Create A transpose
	Matrix3 *Atranspose = new Matrix3();
	this->A->transpose(Atranspose);

	// Multiply Icm * Atranspose
	this->Icm->multiply(Atranspose);

	delete Atranspose;

}
// Given a box's dimenstions, compute the inverse inertia tensor from the 
// center of mass
void Object3d::createInertiaTensor(Matrix3 * I, float a, float b, float c, float M) {
	
	// From http://www.gamedev.net/community/forums/topic.asp?topic_id=57001
	I->set(0,1/(((b*b + c*c) / 3)*M));
	I->set(4,1/(((a*a + c*c) / 3)*M));
	I->set(8,1/(((a*a + b*b) / 3)*M));

	// DEBUG: May need to take the inverse of this diagonal matrix
	// instead! -- Yep, I did take the inverse
}

/*
void Object3d::setOrientation() {

	// Get X,Y,Z rotations from A
	float xr, yr, zr;
	A->getRotationAngles(&xr, &yr, &zr);
	dbRotateObject(id, xr, yr, zr);
	
	// Apply them

	//dbRotateObject(id, o[1], o[2], o[3]);
	//debugShowStr();
}
*/

void Object3d::debugShowStr() {
		char textStr [50];
		sprintf(textStr,"Ang:%f  %f %f",o[0], o[1],o[2]);
		dbText(30,60,textStr);
}
